#include "W25Q.h"

void W25Q_WaitWhileBusy();

void W25Q_BeginCommand(uint8_t cmd) {
	W25Q_EnableSpi();
	W25Q_SendByte(cmd);
}

void W25Q_SendCommand(uint8_t cmd) {
	W25Q_BeginCommand(cmd);
	W25Q_DisableSpi();
}

uint8_t W25Q_SendCommandReadByte(uint8_t cmd) {
	W25Q_BeginCommand(cmd);

	uint8_t output = W25Q_SendByte(0xFF);

	W25Q_DisableSpi();
	return output;
}

void W25Q_Send24BitAddress(uint32_t address) {
	for (int i = 0; i < 3; i++) {
		W25Q_SendByte(address >> ((2 - i) * 8));
	}
}

void W25Q_WriteEnable() {
	W25Q_SendCommand(W25Q_OP_WriteEnable);
}

void W25Q_WriteStatusRegisterVolatileEnable() {
	W25Q_SendCommand(W25Q_OP_VolatileSRWriteEnable);
}

void W25Q_WriteDisable() {
	W25Q_SendCommand(W25Q_OP_WriteDisable);
}

uint8_t W25Q_ReadSR1() {
	return W25Q_SendCommandReadByte(W25Q_OP_ReadSR1);
}

uint8_t W25Q_ReadSR2() {
	return W25Q_SendCommandReadByte(W25Q_OP_ReadSR2);
}

uint8_t W25Q_ReadSR3() {
	return W25Q_SendCommandReadByte(W25Q_OP_ReadSR3);
}

void W25Q_WriteSR(uint8_t opCode, uint8_t newRegValue, uint8_t writeVolatile) {
	if (writeVolatile) {
		W25Q_WriteStatusRegisterVolatileEnable();
	} else {
		W25Q_WriteEnable();
	}
	W25Q_BeginCommand(opCode);
	W25Q_SendByte(newRegValue);
	W25Q_DisableSpi();
	W25Q_WaitWhileBusy();
}

void W25Q_WriteSR1(uint8_t newRegValue) {
	W25Q_WriteSR(W25Q_OP_WriteSR1, newRegValue, 0);
}

void W25Q_WriteSR2(uint8_t newRegValue) {
	W25Q_WriteSR(W25Q_OP_WriteSR2, newRegValue, 0);
}

void W25Q_WriteSR3(uint8_t newRegValue) {
	W25Q_WriteSR(W25Q_OP_WriteSR3, newRegValue, 0);
}

void W25Q_WriteSR1Volatile(uint8_t newRegValue) {
	W25Q_WriteSR(W25Q_OP_WriteSR1, newRegValue, 1);
}

void W25Q_WriteSR2Volatile(uint8_t newRegValue) {
	W25Q_WriteSR(W25Q_OP_WriteSR2, newRegValue, 1);
}

void W25Q_WriteSR3Volatile(uint8_t newRegValue) {
	W25Q_WriteSR(W25Q_OP_WriteSR3, newRegValue, 1);
}

void W25Q_WaitWhileBusy() {
	W25Q_BeginCommand(W25Q_OP_ReadSR1);

	while (W25Q_SendByte(0xFF) & W25Q_REG_SR1_BUSY) {
	}

	W25Q_DisableSpi();
}

uint8_t W25Q_ReadData(uint32_t address, uint8_t* output, uint16_t outputLength) {
	if (address >> 24) { // address cannot have set bits 24-31. It is has only 24bit in length.
		return 1; // argument exception
	}
	uint8_t sr1 = W25Q_ReadSR1();

	if (sr1 & W25Q_REG_SR1_BUSY) {
		return 2; // device is busy and cannot read data
	}

	W25Q_BeginCommand(W25Q_OP_ReadData);
	W25Q_SendByte((uint8_t) (address >> 16));
	W25Q_SendByte((uint8_t) (address >> 8));
	W25Q_SendByte((uint8_t) (address >> 0));

	while (outputLength--) {
		*(output++) = W25Q_SendByte(0xFF);
	}

	W25Q_DisableSpi();
	return 0;
}

uint8_t W25Q_ReadDataFast(uint32_t address, uint8_t* output, uint16_t outputLength) {
	if (address >> 24) { // address cannot have set bits 24-31. It is has only 24bit in length.
		return 1; // argument exception
	}
	uint8_t sr1 = W25Q_ReadSR1();

	if (sr1 & W25Q_REG_SR1_BUSY) {
		return 2; // device is busy and cannot read data
	}

	W25Q_BeginCommand(W25Q_OP_FastRead);
	W25Q_SendByte((uint8_t) (address >> 16));
	W25Q_SendByte((uint8_t) (address >> 8));
	W25Q_SendByte((uint8_t) (address >> 0));

	W25Q_SendByte(0xFF); // dummy byte required for fast read

	while (outputLength--) {
		*(output++) = W25Q_SendByte(0xFF);
	}

	W25Q_DisableSpi();
	return 0;
}

uint8_t W25Q_VerifyData(uint32_t address, uint8_t* data, uint16_t dataLength) {
	if (address >> 24) { // address cannot have set bits 24-31. It is has only 24bit in length.
		return 1; // argument exception
	}
	uint8_t sr1 = W25Q_ReadSR1();

	if (sr1 & W25Q_REG_SR1_BUSY) {
		return 2; // device is busy and cannot read data
	}

	W25Q_BeginCommand(W25Q_OP_FastRead);
	W25Q_SendByte((uint8_t) (address >> 16));
	W25Q_SendByte((uint8_t) (address >> 8));
	W25Q_SendByte((uint8_t) (address >> 0));

	W25Q_SendByte(0xFF); // dummy byte required for fast read

	while (dataLength--) {
		if (*(data++) != W25Q_SendByte(0xFF)) {
			return 3;
		}
	}

	W25Q_DisableSpi();
	return 0;
}

uint8_t W25Q_ProgramPage(uint32_t address, uint8_t* data, uint16_t dataLenth) {
	if (address >> 24) {
		return 1;
	}
	uint8_t sr1 = W25Q_ReadSR1();

	if (sr1 & W25Q_REG_SR1_BUSY) {
		return 2; // device is busy and cannot read data
	}

	if (!(sr1 & W25Q_REG_SR1_WEL)) {
		W25Q_WriteEnable();

		if (!(W25Q_ReadSR1() & W25Q_REG_SR1_WEL)) {
			return 3; // cannot enable writing
		}
	}

	W25Q_BeginCommand(W25Q_OP_PageProgram);
	W25Q_SendByte((uint8_t) (address >> 16));
	W25Q_SendByte((uint8_t) (address >> 8));
	W25Q_SendByte((uint8_t) (address >> 0));

	while (dataLenth--) {
		W25Q_SendByte(*(data++));
	}

	W25Q_DisableSpi();
	W25Q_WaitWhileBusy();
	return 0;
}

uint8_t W25Q_EraseSector(uint32_t address) {
	if (address >> 24) {
		return 1;
	}
	uint8_t sr1 = W25Q_ReadSR1();

	if (sr1 & W25Q_REG_SR1_BUSY) {
		return 2; // device is busy and cannot read data
	}

	if (!(sr1 & W25Q_REG_SR1_WEL)) {
		W25Q_WriteEnable();

		if (!(W25Q_ReadSR1() & W25Q_REG_SR1_WEL)) {
			return 3; // cannot enable writing
		}
	}

	W25Q_BeginCommand(W25Q_OP_SectorErase4K);
	W25Q_SendByte((uint8_t) (address >> 16));
	W25Q_SendByte((uint8_t) (address >> 8));
	W25Q_SendByte((uint8_t) (address >> 0));

	W25Q_DisableSpi();
	W25Q_WaitWhileBusy();
	return 0;
}

uint8_t W25Q_EraseChip() {
	uint8_t sr1 = W25Q_ReadSR1();

	if (sr1 & W25Q_REG_SR1_BUSY) {
		return 2; // device is busy and cannot erase chip
	}

	if (!(sr1 & W25Q_REG_SR1_WEL)) {
		W25Q_WriteEnable();

		if (!(W25Q_ReadSR1() & W25Q_REG_SR1_WEL)) {
			return 3; // cannot enable writing
		}
	}

	W25Q_SendCommand(W25Q_OP_ChipErase);
	W25Q_WaitWhileBusy();
	return 0;
}

void W25Q_ReadManufacturerId(uint8_t* vendorIdOutput, uint8_t* deviceIdOutput) {
	W25Q_BeginCommand(W25Q_OP_DeviceId);
	W25Q_SendByte(0x00);
	W25Q_SendByte(0x00);
	W25Q_SendByte(0x00);

	*vendorIdOutput = W25Q_SendByte(0xFF);
	*deviceIdOutput = W25Q_SendByte(0xFF);

	W25Q_DisableSpi();
}

uint64_t W25Q_ReadUniqueId() {
	uint64_t output = 0;

	W25Q_BeginCommand(W25Q_OP_ReadUniqueId);
	W25Q_SendByte(0xFF);
	W25Q_SendByte(0xFF);
	W25Q_SendByte(0xFF);
	W25Q_SendByte(0xFF);

	for (uint8_t i = 0; i < 8; i++) {
		output |= (((uint64_t) (W25Q_SendByte(0xFF))) << ((7 - i) * 8));
	}

	W25Q_DisableSpi();
	return output;
}

uint32_t W25Q_ReadJedecId() {
	W25Q_BeginCommand(W25Q_OP_JedecId);
	uint32_t output = 0;

	for (uint8_t i = 0; i < 3; i++) {
		output |= (((uint32_t) (W25Q_SendByte(0xFF))) << ((2 - i) * 8));
	}

	W25Q_DisableSpi();
	return output;
}

uint8_t W25Q_ReadSfdpHeader(SfdpHeader* output, uint16_t sfdpHeaderCapacity) {
	W25Q_BeginCommand(W25Q_OP_ReadSfdpRegister);
	for (int i = 0; i < 1 + 3; i++) { // 1 dummy bytes + 3 address bytes
		W25Q_SendByte(0x00); // dummy byte
	}

	if (sfdpHeaderCapacity < 8) {
		W25Q_DisableSpi();
		return 1;
	}

	uint8_t* outputPtr = (uint8_t*) output;

	// load signature
	for (int i = 0; i < 4; i++) {
		sfdpHeaderCapacity--;
		*(outputPtr++) = W25Q_SendByte(0xFF);
	}

	if (output->signature != 0x50444653) {
		W25Q_DisableSpi();
		return 2;
	}

	// load header
	for (int i = 0; i < 4; i++) {
		sfdpHeaderCapacity--;
		*(outputPtr++) = W25Q_SendByte(0xFF);
	}

	// load parameters
	uint16_t bytesToLoad = (output->parametersCount + 1) * 8;
	uint8_t status = 0;
	if (bytesToLoad > sfdpHeaderCapacity) {
		status = 3;
		bytesToLoad = sfdpHeaderCapacity;
	}

	while (bytesToLoad--) {
		*(outputPtr++) = W25Q_SendByte(0xFF);
	}

	W25Q_DisableSpi();
	return status;
}

void W25Q_ReadSfdpParam(SfdpParameterHeader* param, uint8_t* output) {
	W25Q_BeginCommand(W25Q_OP_ReadSfdpRegister);

	// send sfdp offset
	for (int i = 0; i < 3; i++) { // 1 dummy bytes + 3 address bytes
		W25Q_SendByte(param->ptrBytes[2 - i]); // dummy byte
	}

	W25Q_SendByte(0xFF); // dummy byte

	uint16_t remainingBytes = param->length * 4; // length is in double words
	while (remainingBytes--) {
		*(output++) = W25Q_SendByte(0xFF);
	}

	W25Q_DisableSpi();
}

uint8_t W25Q_GetCurrentProtectionMode() {
	if (W25Q_ReadSR3() & W25Q_REG_SR3_WPS) {
		return W25Q_PROTECTION_MODE_INDIVIDUAL_BLOCK_LOCK;
	} else {
		return W25Q_PROTECTION_MODE_AREA_LOCK;
	}
}

uint8_t W25Q_SetCurrentProtectionMode(uint8_t newMode, uint8_t writeVolatile) {
	if (newMode != W25Q_PROTECTION_MODE_AREA_LOCK && newMode != W25Q_PROTECTION_MODE_INDIVIDUAL_BLOCK_LOCK) {
		return 1;
	}

	uint8_t sr3 = W25Q_ReadSR3();
	uint8_t originalSr3 = sr3;
	if (newMode == W25Q_PROTECTION_MODE_AREA_LOCK) {
		sr3 |= W25Q_REG_SR3_WPS;
	} else {
		sr3 &= ~W25Q_REG_SR3_WPS;
	}

	if (sr3 != originalSr3) {
		if (writeVolatile) {
			W25Q_WriteSR3Volatile(sr3);
		} else {
			W25Q_WriteSR3(sr3);
		}
	}

	return W25Q_GetCurrentProtectionMode() != newMode;
}

uint8_t W25Q_GetDriverStrength() {
	uint8_t regValue = (W25Q_ReadSR3() & 0x60) >> 5;
	switch (regValue) {
	case 0:
		return W25Q_DRIVER_STRENGTH_100;
		break;
	case 1:
		return W25Q_DRIVER_STRENGTH_75;
		break;
	case 2:
		return W25Q_DRIVER_STRENGTH_50;
		break;
	case 3:
		return W25Q_DRIVER_STRENGTH_25;
		break;
	}
	return 0xFF; // cannot happen
}

uint8_t W25Q_SetDriverStrength(uint8_t newStrength, uint8_t writeVolatile) {
	if (newStrength != W25Q_DRIVER_STRENGTH_100 && newStrength != W25Q_DRIVER_STRENGTH_75
			&& newStrength != W25Q_DRIVER_STRENGTH_50 && newStrength != W25Q_DRIVER_STRENGTH_25) {
		return 1;
	}

	uint8_t sr3 = W25Q_ReadSR3();
	uint8_t originalSr3 = sr3;
	if (newStrength == W25Q_DRIVER_STRENGTH_100) {
		sr3 &= ~W25Q_REG_SR3_DRV0;
		sr3 &= ~W25Q_REG_SR3_DRV1;
	} else if (newStrength == W25Q_DRIVER_STRENGTH_75) {
		sr3 |= W25Q_REG_SR3_DRV0;
		sr3 &= ~W25Q_REG_SR3_DRV1;
	} else if (newStrength == W25Q_DRIVER_STRENGTH_50) {
		sr3 &= ~W25Q_REG_SR3_DRV0;
		sr3 |= W25Q_REG_SR3_DRV1;
	} else {
		sr3 |= W25Q_REG_SR3_DRV0;
		sr3 |= W25Q_REG_SR3_DRV1;
	}

	if (sr3 != originalSr3) {
		if (writeVolatile) {
			W25Q_WriteSR3Volatile(sr3);
		} else {
			W25Q_WriteSR3(sr3);
		}
	}

	return W25Q_GetDriverStrength() != newStrength;
}

uint8_t W25Q_GetHoldResetPinMode() {
	if (W25Q_ReadSR3() & W25Q_REG_SR3_HOLD_RST) {
		return W25Q_HOLD_RESET_MODE_RESET;
	} else {
		return W25Q_HOLD_RESET_MODE_HOLD;
	}
}

uint8_t W25Q_SetHoldResetPinMode(uint8_t newMode, uint8_t writeVolatile) {
	if (newMode != W25Q_HOLD_RESET_MODE_HOLD && newMode != W25Q_HOLD_RESET_MODE_RESET) {
		return 1;
	}

	uint8_t sr3 = W25Q_ReadSR3();
	uint8_t originalSr3 = sr3;
	if (newMode == W25Q_HOLD_RESET_MODE_RESET) {
		sr3 |= W25Q_REG_SR3_HOLD_RST;
	} else {
		sr3 &= ~W25Q_REG_SR3_HOLD_RST;
	}

	if (sr3 != originalSr3) {
		if (writeVolatile) {
			W25Q_WriteSR3Volatile(sr3);
		} else {
			W25Q_WriteSR3(sr3);
		}
	}

	return W25Q_GetHoldResetPinMode() != newMode;
}

uint8_t W25Q_GetCapacity(uint64_t* output) {
	uint8_t buffer[16]; // 16 bytes is SFDP header + 1 param.
	SfdpHeader* header = (SfdpHeader*) buffer;
	uint8_t status = W25Q_ReadSfdpHeader(header, sizeof(buffer));
	if (status != 0 && status != 3) { // 3 is OK because we do not need more than 1 parameter. So we do not need to pass bigger buffer.
		return 1;
	}

	// we need to load only 2 bytes of header.
	if (header->parameters[0].length < 2) {
		return 2;
	}
	header->parameters[0].length = 2;

	SfdpBasicFlashParameters basicParams;
	W25Q_ReadSfdpParam(header->parameters + 0, (uint8_t*) &basicParams);

	if (basicParams.density & 0x80000000) {
		*output = (2 << (basicParams.density & 0x7FFFFFFF)) / 8;
		if (*output == 0) {
			return 4;
		}
	} else {
		*output = ((uint64_t) basicParams.density + 1) / 8;
	}

	return 0;
}

uint8_t W25Q_ProgramSecurityRegister(uint32_t address, uint8_t* data, uint16_t length) {
	if (length > 256) {
		return 1;
	}

	W25Q_WriteEnable();

	W25Q_BeginCommand(W25Q_OP_ProgramSecurityRegister);
	W25Q_Send24BitAddress(address);

	while (length--) {
		W25Q_SendByte(*(data++));
	}

	W25Q_DisableSpi();
	W25Q_WaitWhileBusy();
	return 0;
}

void W25Q_EraseSecurityRegister(uint32_t address) {
	W25Q_WriteEnable();

	W25Q_BeginCommand(W25Q_OP_EraseSecurityRegister);
	W25Q_Send24BitAddress(address);

	W25Q_DisableSpi();
	W25Q_WaitWhileBusy();
}

uint8_t W25Q_ReadSecurityRegister(uint32_t address, uint8_t* output, uint16_t outputLength) {
	if (outputLength > 256) {
		return 1;
	}

	W25Q_BeginCommand(W25Q_OP_ReadSecurityRegister);
	W25Q_Send24BitAddress(address);

	W25Q_SendByte(0xFF);

	while (outputLength--) {
		*(output++) = W25Q_SendByte(0xFF);
	}

	W25Q_DisableSpi();
	return 0;
}

void W25Q_EnterQpiMode() {
	W25Q_SendCommand(W25Q_OP_EnterQpiMode);
}

void W25Q_PowerDown() {
	W25Q_SendCommand(W25Q_OP_PowerDown);
}

void W25Q_ReleasePowerDown() {
	W25Q_SendCommand(W25Q_OP_ReleasePowerDown);
}

void W25Q_LockBlock(uint32_t address) {
	W25Q_BeginCommand(W25Q_OP_IndividualBlockLock);
	W25Q_Send24BitAddress(address);
	W25Q_DisableSpi();
}

void W25Q_UnlockBlock(uint32_t address) {
	W25Q_BeginCommand(W25Q_OP_IndividualBlockUnlock);
	W25Q_Send24BitAddress(address);
	W25Q_DisableSpi();
}

uint8_t W25Q_IsBlockLocked(uint32_t address) {
	W25Q_BeginCommand(W25Q_OP_ReadBlockLock);
	W25Q_Send24BitAddress(address);
	uint8_t output = W25Q_SendByte(0xFF);
	W25Q_DisableSpi();
	return output & 0x01;
}

void W25Q_LockAllBLocks() {
	W25Q_SendCommand(W25Q_OP_GlobalBlockLock);
}

void W25Q_UnlockAllBlocks() {
	W25Q_SendCommand(W25Q_OP_GlobalBlockUnlock);
}

void W25Q_Reset() {
	W25Q_SendCommand(W25Q_OP_EnableReset);
	W25Q_SendCommand(W25Q_OP_Reset);
}

void W25Q_GetArrayLock(uint8_t* sec, uint8_t* tb, uint8_t* bp, uint8_t* cmp) {
	uint8_t sr1 = W25Q_ReadSR1();
	if (sec != NULL) {
		*sec = !!(sr1 & W25Q_REG_SR1_SEC);
	}
	if (tb != NULL) {
		*tb = !!(sr1 & W25Q_REG_SR1_TB);
	}
	if (bp != NULL) {
		*bp = (sr1 & (W25Q_REG_SR1_BP0 | W25Q_REG_SR1_BP1 | W25Q_REG_SR1_BP2)) >> 2;
	}
	if (cmp != NULL) {
		*cmp = !!(W25Q_ReadSR2() & W25Q_REG_SR2_CMP);
	}
}

void W25Q_SetArrayLock(uint8_t sec, uint8_t tb, uint8_t bp, uint8_t cmp, uint8_t writeVolatile) {
	uint8_t sr1 = W25Q_ReadSR1();
	uint8_t sr2 = W25Q_ReadSR2();
	uint8_t sr1Orig = sr1;
	uint8_t sr2Orig = sr2;

	if (sec) {
		sr1 |= W25Q_REG_SR1_SEC;
	} else {
		sr1 &= ~W25Q_REG_SR1_SEC;
	}
	if (tb) {
		sr1 |= W25Q_REG_SR1_TB;
	} else {
		sr1 &= ~W25Q_REG_SR1_TB;
	}
	if (bp & 0x01) {
		sr1 |= W25Q_REG_SR1_BP0;
	} else {
		sr1 &= ~W25Q_REG_SR1_BP0;
	}
	if (bp & 0x02) {
		sr1 |= W25Q_REG_SR1_BP1;
	} else {
		sr1 &= ~W25Q_REG_SR1_BP1;
	}
	if (bp & 0x04) {
		sr1 |= W25Q_REG_SR1_BP2;
	} else {
		sr1 &= ~W25Q_REG_SR1_BP2;
	}

	if (cmp) {
		sr2 |= W25Q_REG_SR2_CMP;
	} else {
		sr2 &= ~W25Q_REG_SR2_CMP;
	}

	if (sr1 != sr1Orig) {
		if (writeVolatile) {
			W25Q_WriteSR1Volatile(sr1);
		} else {
			W25Q_WriteSR1(sr1);
		}
	}

	if (sr2 != sr2Orig) {
		if (writeVolatile) {
			W25Q_WriteSR2Volatile(sr2);
		} else {
			W25Q_WriteSR2(sr2);
		}
	}
}

uint8_t W25Q_GetStatusRegisterProtection() {
	uint8_t sr1 = W25Q_ReadSR1();
	uint8_t sr2 = W25Q_ReadSR2();

	uint8_t mode = ((!!(sr2 & W25Q_REG_SR2_SRP1)) << 1) | (!!(sr1 & W25Q_REG_SR1_SRP0));

	switch (mode) {
	case 0:
		return W25Q_STATUS_REGISTER_PROTECTION_SOFTWARE;
		break;
	case 1:
		return W25Q_STATUS_REGISTER_PROTECTION_HARDWARE;
		break;
	case 2:
		return W25Q_STATUS_REGISTER_PROTECTION_POWER_SUPLY_LOCKDOWN;
		break;
	case 3:
		return W25Q_STATUS_REGISTER_PROTECTION_ONE_TIME_PROGRAM;
		break;
	}

	return 0xFF; // error, could not happen
}

uint8_t W25Q_SetStatusRegisterProtection(uint8_t newProtection, uint8_t writeVolatile) {
	if (newProtection != W25Q_STATUS_REGISTER_PROTECTION_SOFTWARE
			&& newProtection != W25Q_STATUS_REGISTER_PROTECTION_HARDWARE
			&& newProtection != W25Q_STATUS_REGISTER_PROTECTION_POWER_SUPLY_LOCKDOWN
			&& newProtection != W25Q_STATUS_REGISTER_PROTECTION_ONE_TIME_PROGRAM) {
		return 1;
	}

	uint8_t sr1 = W25Q_ReadSR1();
	uint8_t sr2 = W25Q_ReadSR2();
	uint8_t sr1Orig = sr1;
	uint8_t sr2Orig = sr2;

	if (newProtection & 0x01) {
		sr1 |= W25Q_REG_SR1_SRP0;
	} else {
		sr1 &= ~W25Q_REG_SR1_SRP0;
	}

	if (newProtection & 0x02) {
		sr2 |= W25Q_REG_SR2_SRP1;
	} else {
		sr2 &= ~W25Q_REG_SR2_SRP1;
	}

	if (sr1 != sr1Orig) {
		if (writeVolatile) {
			W25Q_WriteSR1Volatile(sr1);
		} else {
			W25Q_WriteSR1(sr1);
		}
	}

	if (sr2 != sr2Orig) {
		if (writeVolatile) {
			W25Q_WriteSR2Volatile(sr2);
		} else {
			W25Q_WriteSR2(sr2);
		}
	}

	return W25Q_GetStatusRegisterProtection() != newProtection;
}

uint8_t W25Q_EnableQuadMode(uint8_t writeVolatile) {
	uint8_t sr2 = W25Q_ReadSR2();
	uint8_t origSr2 = sr2;

	sr2 |= W25Q_REG_SR2_QE;

	if (sr2 != origSr2) {
		if (writeVolatile) {
			W25Q_WriteSR2Volatile(sr2);
		} else {
			W25Q_WriteSR2(sr2);
		}
	}

	return !W25Q_IsQuadModeEnabled();
}

uint8_t W25Q_DisableQuadMode(uint8_t writeVolatile) {
	uint8_t sr2 = W25Q_ReadSR2();
	uint8_t origSr2 = sr2;

	sr2 &= ~W25Q_REG_SR2_QE;

	if (sr2 != origSr2) {
		if (writeVolatile) {
			W25Q_WriteSR2Volatile(sr2);
		} else {
			W25Q_WriteSR2(sr2);
		}
	}

	return W25Q_IsQuadModeEnabled();
}

uint8_t W25Q_IsQuadModeEnabled() {
	return !!(W25Q_ReadSR2() & W25Q_REG_SR2_QE);
}

uint8_t W25Q_LockSecurityRegister(uint8_t registerNum) {
	if (registerNum < 1 || registerNum > 3) {
		return 1;
	}

	uint8_t sr2 = W25Q_ReadSR2();
	uint8_t origSr2 = sr2;

	sr2 |= (1 << (2 + registerNum));

	if (sr2 == origSr2) {
		return 2;
	}

	W25Q_WriteSR2(sr2);

	return !W25Q_IsSecurityRegisterLocked(registerNum);
}

uint8_t W25Q_IsSecurityRegisterLocked(uint8_t registerNum) {
	if (registerNum < 1 || registerNum > 3) {
		return 0; // invalid register number is marked as unlocked.
	}

	return !!(W25Q_ReadSR2() & (1 << (2 + registerNum)));
}
