#include <avr/io.h>
#include <avr/interrupt.h>

#define F_CPU 24000000UL
#include <util/delay.h>

#include <math.h>
#include <stddef.h>
#include <stdint.h>
#include <ctype.h>
#include <string.h>

#include "W25Q.h"
#include "LED.h"

#define AUDIO_TCB				TCB0
#define AUDIO_TCB_INT_vect		TCB0_INT_vect
#define AUDIO_TCB_INT_vect_num	TCB0_INT_vect_num

static uint16_t audioBuffer[512];
static uint16_t audioBufferLen = 512;
static uint16_t audioBufferIndex = 0;

static uint8_t loadAudioBufferRequested = 0;
static uint8_t audioBufferToLoad = 0;

static uint32_t flashAddress = 44;
static uint32_t soundLength = 0;

ISR(AUDIO_TCB_INT_vect) {
	PORTD.OUTSET = PIN7_bm;
	
	AUDIO_TCB.INTFLAGS = TCB_CAPT_bm;
	
	// xor 0x8000 converts int16_t to uint16_t while correcting negative values.
	// efficiently it is addition of 32768
	// AVR DAC has register value left-aligned. We do not need to shift it manually. 
	// 6-LSBs are dropped at register write.
	DAC0.DATA = audioBuffer[audioBufferIndex++] ^ 0x8000;
	// DAC0.DATA = (audioBuffer[audioBufferIndex++] ^ 0x8000) >> 1;
	
	if (audioBufferIndex == audioBufferLen) {
		audioBufferIndex = 0;
	}
	
	// following operation is bit optimized
	// equivalent condition is (audioBufferIndex == 256 || audioBufferIndex == 0)
	if ((audioBufferIndex & 0xFF) == 0) {
		loadAudioBufferRequested = 1;
	}
	
	LED_Show();
	
	PORTD.OUTCLR = PIN7_bm;
}

void Audio_Init() {
	VREF.DAC0REF = VREF_REFSEL_1V024_gc;
	DAC0.CTRLA = DAC_OUTEN_bm | DAC_ENABLE_bm;
	
	PORTD.OUTSET = PIN6_bm;
	PORTD.OUTCLR = PIN6_bm;
	
	AUDIO_TCB.INTCTRL = TCB_CAPT_bm;
	AUDIO_TCB.CCMP = 272; // F_CPU / PDIV / sampleFreq (24MHz; PDIV=2; freq=44100 => CCMP=272.109)
	// AUDIO_TCB.CCMP = 227; // F_CPU / PDIV / sampleFreq (20MHz; PDIV=2; freq=44100 => CCMP=226.757)
	AUDIO_TCB.CTRLA = TCB_ENABLE_bm;
}

void Clock_Init() {
	CCP = CCP_IOREG_gc;
	CLKCTRL.OSCHFCTRLA = CLKCTRL_FRQSEL_24M_gc;
	CCP = CCP_IOREG_gc;
	CLKCTRL.MCLKCTRLB = CLKCTRL_PDIV_2X_gc | CLKCTRL_PEN_bm;
	CCP = CCP_IOREG_gc;
	CLKCTRL.MCLKCTRLA = CLKCTRL_CLKSEL_OSCHF_gc;
	
	for (volatile int i = 0; i < 1000; i++) {}
}

void LoadBuffer() {
	if (audioBufferToLoad) {
		audioBufferToLoad = 0;
		W25Q_ReadData(flashAddress, (uint8_t*)(audioBuffer + 256), 256 * sizeof(int16_t));
	} else {
		audioBufferToLoad = 1;
		W25Q_ReadData(flashAddress, (uint8_t*)(audioBuffer), 256 * sizeof(int16_t));
	}
	flashAddress += 256 * sizeof(int16_t);
	if (flashAddress > soundLength + 44) {
		flashAddress = 44;
	}
}

void LoadHeader() {
	uint8_t header[44];
	W25Q_ReadData(0, header, sizeof(header));
	// check invalid or unsupported WAV file
	if (strncmp((char*)(header + 0x00), "RIFF", 4) != 0 ||
		strncmp((char*)(header + 0x08), "WAVEfmt", 7) != 0 ||
		strncmp((char*)(header + 0x24), "data", 4) != 0 ||
		*(uint32_t*)(header + 0x10) != 16 ||
		*(uint16_t*)(header + 0x14) != 1 ||
		*(uint16_t*)(header + 0x16) != 1 ||
		*(uint32_t*)(header + 0x18) != 44100 ||
		*(uint32_t*)(header + 0x1C) != 44100 * sizeof(int16_t) ||
		*(uint16_t*)(header + 0x20) != 2 ||
		0) {
			
		for (int i = 0; i < 5; i++) {
			LED_PORT.OUTSET = PIN3_bm;
			_delay_ms(250);
			LED_PORT.OUTCLR = PIN3_bm;
			_delay_ms(250);
		}
		_delay_ms(1000);
		
		CCP = CCP_IOREG_gc;
		RSTCTRL.SWRR = RSTCTRL_SWRST_bm;
			
		
	}
	soundLength = *(uint32_t*)(header + 0x28);
}

int main(void) {
	Clock_Init();
	//UART_Init();
	W25Q_Init();
	LED_Init();
	
	LoadHeader();
	//soundLength = 1024;
	LoadBuffer();
	LoadBuffer();
	
	Audio_Init();
	
	PORTD.DIRSET = PIN7_bm;
	
	
	sei();
	
	uint8_t updateLedSkip = 0;
	
	while (1) {
		if (loadAudioBufferRequested) {
			loadAudioBufferRequested = 0;
			LoadBuffer();
			
			if ((updateLedSkip++ & 0x1) == 0) {
				LED_Animate();
			}
			
		}
	}
	
}

